"""djpress URLs file."""

from django.urls import path, re_path

from djpress.conf import settings
from djpress.feeds import PostFeed
from djpress.views import (
    archives_posts,
    author_posts,
    category_posts,
    index,
    post_detail,
)

regex_path = r"^(?P<path>[0-9A-Za-z/_-]*)$"
regex_archives = r"(?P<year>\d{4})(?:/(?P<month>\d{2})(?:/(?P<day>\d{2}))?)?$"

if settings.APPEND_SLASH:
    regex_path = regex_path[:-1] + "/$"
    regex_archives = regex_archives[:-1] + "/$"

app_name = "djpress"

urlpatterns = []

if settings.CATEGORY_PATH_ENABLED and settings.CATEGORY_PATH:
    urlpatterns += [
        path(
            f"{settings.CATEGORY_PATH}/<slug:slug>/",
            category_posts,
            name="category_posts",
        ),
    ]

if settings.AUTHOR_PATH_ENABLED:
    urlpatterns += [
        path(
            f"{settings.AUTHOR_PATH}/<str:author>/",
            author_posts,
            name="author_posts",
        ),
    ]

if settings.ARCHIVES_PATH_ENABLED and settings.ARCHIVES_PATH:
    urlpatterns += [
        re_path(
            settings.ARCHIVES_PATH + "/" + regex_archives,
            archives_posts,
            name="archives_posts",
        ),
    ]

if settings.RSS_ENABLED and settings.RSS_PATH:
    urlpatterns += [
        path(
            f"{settings.RSS_PATH}/",
            PostFeed(),
            name="rss_feed",
        ),
    ]

urlpatterns += [
    path("", index, name="index"),
    re_path(
        regex_path,
        post_detail,
        name="post_detail",
    ),
]



import pytest

from django.contrib.auth.models import User

from djpress.conf import settings
from djpress.models import Category, Post
from djpress.templatetags.helpers import (
    categories_html,
    category_link,
    post_read_more_link,
)


@pytest.fixture
def user():
    user = User.objects.create_user(
        username="testuser",
        password="testpass",
        first_name="Test",
        last_name="User",
    )
    return user


@pytest.fixture
def category1():
    category = Category.objects.create(
        name="General",
        slug="general",
    )
    return category


@pytest.fixture
def category2():
    category = Category.objects.create(
        name="News",
        slug="news",
    )
    return category


@pytest.fixture
def category3():
    category = Category.objects.create(
        name="Development",
        slug="dev",
    )
    return category


@pytest.fixture
def test_post(user, category1):
    post = Post.post_objects.create(
        title="Test Post",
        slug="test-post",
        content="This is a test post.",
        author=user,
        status="published",
        post_type="post",
    )
    post.categories.set([category1])
    return post


@pytest.mark.django_db
def test_categories_html(category1, category2, category3):
    assert settings.CATEGORY_PATH == "test-url-category"

    categories = Category.objects.all()
    assert list(categories) == [category1, category2, category3]

    # Test case 1
    outer = "ul"
    outer_class = "categories"
    link_class = "category"
    expected_output = (
        '<ul class="categories">'
        f'<li><a href="/{settings.CATEGORY_PATH}/{category1.slug}/" title="View all posts in the {category1.name} category" class="{link_class}">{category1.name}</a></li>'
        f'<li><a href="/{settings.CATEGORY_PATH}/{category2.slug}/" title="View all posts in the {category2.name} category" class="{link_class}">{category2.name}</a></li>'
        f'<li><a href="/{settings.CATEGORY_PATH}/{category3.slug}/" title="View all posts in the {category3.name} category" class="{link_class}">{category3.name}</a></li>'
        "</ul>"
    )
    assert (
        categories_html(
            categories, outer=outer, outer_class=outer_class, link_class=link_class
        )
        == expected_output
    )

    # Test case 2
    outer = "ul"
    outer_class = ""
    link_class = ""
    expected_output = (
        "<ul>"
        f'<li><a href="/{settings.CATEGORY_PATH}/{category1.slug}/" title="View all posts in the {category1.name} category">{category1.name}</a></li>'
        f'<li><a href="/{settings.CATEGORY_PATH}/{category2.slug}/" title="View all posts in the {category2.name} category">{category2.name}</a></li>'
        f'<li><a href="/{settings.CATEGORY_PATH}/{category3.slug}/" title="View all posts in the {category3.name} category">{category3.name}</a></li>'
        "</ul>"
    )
    assert (
        categories_html(
            categories, outer=outer, outer_class=outer_class, link_class=link_class
        )
        == expected_output
    )

    # Test case 3
    outer = "div"
    outer_class = "categories"
    link_class = "category"
    expected_output = (
        '<div class="categories">'
        f'<a href="/{settings.CATEGORY_PATH}/{category1.slug}/" title="View all posts in the {category1.name} category" class="{link_class}">{category1.name}</a>, '
        f'<a href="/{settings.CATEGORY_PATH}/{category2.slug}/" title="View all posts in the {category2.name} category" class="{link_class}">{category2.name}</a>, '
        f'<a href="/{settings.CATEGORY_PATH}/{category3.slug}/" title="View all posts in the {category3.name} category" class="{link_class}">{category3.name}</a>'
        "</div>"
    )
    assert (
        categories_html(
            categories, outer=outer, outer_class=outer_class, link_class=link_class
        )
        == expected_output
    )

    # Test case 4
    outer = "div"
    outer_class = ""
    link_class = ""
    expected_output = (
        "<div>"
        f'<a href="/{settings.CATEGORY_PATH}/{category1.slug}/" title="View all posts in the {category1.name} category">{category1.name}</a>, '
        f'<a href="/{settings.CATEGORY_PATH}/{category2.slug}/" title="View all posts in the {category2.name} category">{category2.name}</a>, '
        f'<a href="/{settings.CATEGORY_PATH}/{category3.slug}/" title="View all posts in the {category3.name} category">{category3.name}</a>'
        "</div>"
    )
    assert (
        categories_html(
            categories, outer=outer, outer_class=outer_class, link_class=link_class
        )
        == expected_output
    )

    # Test case 5
    outer = "span"
    outer_class = "categories"
    link_class = "category"
    expected_output = (
        '<span class="categories">'
        f'<a href="/{settings.CATEGORY_PATH}/{category1.slug}/" title="View all posts in the {category1.name} category" class="{link_class}">{category1.name}</a>, '
        f'<a href="/{settings.CATEGORY_PATH}/{category2.slug}/" title="View all posts in the {category2.name} category" class="{link_class}">{category2.name}</a>, '
        f'<a href="/{settings.CATEGORY_PATH}/{category3.slug}/" title="View all posts in the {category3.name} category" class="{link_class}">{category3.name}</a>'
        "</span>"
    )
    assert (
        categories_html(
            categories, outer=outer, outer_class=outer_class, link_class=link_class
        )
        == expected_output
    )

    # Test case 6
    outer = "span"
    outer_class = ""
    link_class = ""
    expected_output = (
        "<span>"
        f'<a href="/{settings.CATEGORY_PATH}/{category1.slug}/" title="View all posts in the {category1.name} category">{category1.name}</a>, '
        f'<a href="/{settings.CATEGORY_PATH}/{category2.slug}/" title="View all posts in the {category2.name} category">{category2.name}</a>, '
        f'<a href="/{settings.CATEGORY_PATH}/{category3.slug}/" title="View all posts in the {category3.name} category">{category3.name}</a>'
        "</span>"
    )
    assert (
        categories_html(
            categories, outer=outer, outer_class=outer_class, link_class=link_class
        )
        == expected_output
    )


@pytest.mark.django_db
def testcategory_link(category1):
    assert settings.CATEGORY_PATH == "test-url-category"

    # Test case 1 - no link class
    link_class = ""
    expected_output = f'<a href="/{settings.CATEGORY_PATH}/{category1.slug}/" title="View all posts in the {category1.name} category">{category1.name}</a>'
    assert category_link(category1, link_class) == expected_output

    # Test case 2 - with link class
    link_class = "category-class"
    expected_output = f'<a href="/{settings.CATEGORY_PATH}/{category1.slug}/" title="View all posts in the {category1.name} category" class="{link_class}">{category1.name}</a>'
    assert category_link(category1, link_class) == expected_output


@pytest.mark.django_db
def test_post_read_more_link(test_post):
    assert settings.POST_READ_MORE_TEXT == "Test read more..."
    assert settings.POST_PREFIX == "test-posts"

    # Test case 1 - use the app settings for the read more text
    link_class = ""
    read_more_text = ""
    expected_output = f'<p><a href="/{settings.POST_PREFIX}/{test_post.slug}/">{settings.POST_READ_MORE_TEXT}</a></p>'
    assert post_read_more_link(test_post, link_class, read_more_text) == expected_output

    # Test case 2 - use all options
    link_class = "read-more"
    read_more_text = "Continue reading"
    expected_output = f'<p><a href="/{settings.POST_PREFIX}/{test_post.slug}/" class="{link_class}">{read_more_text}</a></p>'
    assert post_read_more_link(test_post, link_class, read_more_text) == expected_output



import pytest
from django.contrib.auth.models import User
from django.http import Http404
from django.utils import timezone
from django.utils.text import slugify

from djpress.conf import settings
from djpress.models import Category, Post


@pytest.fixture
def user():
    return User.objects.create_user(username="testuser", password="testpass")


@pytest.mark.django_db
def test_post_model(user):
    category = Category.objects.create(name="Test Category", slug="test-category")
    post = Post.post_objects.create(
        title="Test Content",
        slug="test-content",
        content="This is a test content.",
        author=user,
        status="published",
        post_type="post",
    )
    post.categories.add(category)
    assert post.title == "Test Content"
    assert post.slug == "test-content"
    assert post.author == user
    assert post.status == "published"
    assert post.post_type == "post"
    assert post.categories.count() == 1
    assert str(post) == "Test Content"


@pytest.mark.django_db
def test_post_methods(user):
    category1 = Category.objects.create(name="Category 1", slug="category-1")
    category2 = Category.objects.create(name="Category 2", slug="category-2")

    Post.post_objects.create(
        title="Test Post 1",
        slug="test-post-1",
        content="This is test post 1.",
        author=user,
        status="published",
        post_type="post",
    ).categories.add(category1)

    Post.post_objects.create(
        title="Test Post 2",
        slug="test-post-2",
        content="This is test post 2.",
        author=user,
        status="draft",
        post_type="post",
    )

    assert Post.post_objects.all().count() == 2
    assert (
        Post.post_objects.get_published_post_by_slug("test-post-1").title
        == "Test Post 1"
    )
    assert Post.post_objects.get_published_posts_by_category(category1).count() == 1
    assert Post.post_objects.get_published_posts_by_category(category2).count() == 0


@pytest.mark.django_db
def test_get_published_content_with_future_date(user):
    Post.post_objects.create(
        title="Past Post",
        slug="past-post",
        content="This is a past post.",
        author=user,
        status="published",
        post_type="post",
        date=timezone.now() - timezone.timedelta(days=1),
    )
    Post.post_objects.create(
        title="Future Post",
        slug="future-post",
        content="This is a future post.",
        author=user,
        status="published",
        post_type="post",
        date=timezone.now() + timezone.timedelta(days=1),
    )
    assert Post.post_objects.all().count() == 2
    assert Post.post_objects.get_published_posts().count() == 1


@pytest.mark.django_db
def test_get_published_content_ordering(user):
    Post.post_objects.create(
        title="Older Post",
        slug="older-post",
        content="This is an older post.",
        author=user,
        status="published",
        post_type="post",
        date=timezone.now() - timezone.timedelta(days=2),
    )
    Post.post_objects.create(
        title="Newer Post",
        slug="newer-post",
        content="This is a newer post.",
        author=user,
        status="published",
        post_type="post",
        date=timezone.now() - timezone.timedelta(days=1),
    )
    posts = Post.post_objects.all()
    assert posts[0].title == "Newer Post"
    assert posts[1].title == "Older Post"


@pytest.mark.django_db
def test_get_published_post_by_slug_with_future_date(user):
    Post.post_objects.create(
        title="Future Post",
        slug="future-post",
        content="This is a future post.",
        author=user,
        status="published",
        post_type="post",
        date=timezone.now() + timezone.timedelta(days=1),
    )
    with pytest.raises(ValueError):
        Post.post_objects.get_published_post_by_slug("future-post")


@pytest.mark.django_db
def test_get_published_content_by_category_with_future_date(user):
    category = Category.objects.create(name="Test Category", slug="test-category")
    Post.post_objects.create(
        title="Past Post",
        slug="past-post",
        content="This is a past post.",
        author=user,
        status="published",
        post_type="post",
        date=timezone.now() - timezone.timedelta(days=1),
    ).categories.add(category)
    Post.post_objects.create(
        title="Future Post",
        slug="future-post",
        content="This is a future post.",
        author=user,
        status="published",
        post_type="post",
        date=timezone.now() + timezone.timedelta(days=1),
    ).categories.add(category)
    assert Post.post_objects.get_published_posts_by_category(category).count() == 1


@pytest.mark.django_db
def test_post_slug_generation(user):
    # Test case 1: Slug generated from title
    post1 = Post.post_objects.create(
        title="My First Blog Post",
        content="This is the content of my first blog post.",
        author=user,
    )
    assert post1.slug == "my-first-blog-post"

    # Test case 2: Slug not overridden when provided
    post2 = Post.post_objects.create(
        title="My Second Blog Post",
        slug="custom-slug",
        content="This is the content of my second blog post.",
        author=user,
    )
    assert post2.slug == "custom-slug"

    # Test case 3: Slug generated with special characters
    post3 = Post.post_objects.create(
        title="My Third Blog Post!",
        content="This is the content of my third blog post.",
        author=user,
    )
    assert post3.slug == "my-third-blog-post"

    # Test case 4: Slug generated with non-ASCII characters
    post4 = Post.post_objects.create(
        title="My Post with 😊 Emoji",
        content="This is the content of the post with an emoji in the title.",
        author=user,
    )
    assert post4.slug == "my-post-with-emoji"

    # Test case 5: Raise error for invalid title
    with pytest.raises(ValueError) as exc_info:
        Post.post_objects.create(
            title="!@#$%^&*()",
            content="This is the content of the post with an invalid title.",
            author=user,
        )
    assert str(exc_info.value) == "Invalid title. Unable to generate a valid slug."


@pytest.mark.django_db
def test_post_markdown_rendering(user):
    assert settings.MARKDOWN_EXTENSIONS == ["fenced_code", "codehilite", "tables"]

    # Test case 1: Render markdown with basic formatting
    post1 = Post.post_objects.create(
        title="Post with Markdown",
        content="# Heading\n\nThis is a paragraph with **bold** and *italic* text.",
        author=user,
    )
    expected_html = "<h1>Heading</h1>\n<p>This is a paragraph with <strong>bold</strong> and <em>italic</em> text.</p>"
    assert post1.content_markdown == expected_html

    # Test case 2: Render markdown with code block
    post2 = Post.post_objects.create(
        title="Post with Code Block",
        content='```python\nprint("Hello, World!")\n```',
        author=user,
    )
    expected_html = '<div class="codehilite"><pre><span></span><code><span class="nb">print</span><span class="p">(</span><span class="s2">&quot;Hello, World!&quot;</span><span class="p">)</span>\n</code></pre></div>'
    assert post2.content_markdown == expected_html

    # Test case 3: Render markdown with fenced code block
    post3 = Post.post_objects.create(
        title="Post with Fenced Code Block",
        content="```\nThis is a fenced code block.\n```",
        author=user,
    )
    expected_html = '<div class="codehilite"><pre><span></span><code>This is a fenced code block.\n</code></pre></div>'
    assert post3.content_markdown == expected_html


@pytest.mark.django_db
def test_post_truncated_content_markdown(user):
    # Confirm the truncate tag is set according to settings_testing.py
    truncate_tag = "<!--test-more-->"
    assert settings.TRUNCATE_TAG == truncate_tag

    # Test case 1: Content with "read more" tag
    post1 = Post.post_objects.create(
        title="Post with Read More",
        content=f"This is the intro.\n\n{truncate_tag}\n\nThis is the rest of the content.",
        author=user,
    )
    expected_truncated_content = "<p>This is the intro.</p>"
    assert post1.truncated_content_markdown == expected_truncated_content

    # Test case 2: Content without "read more" tag
    post2 = Post.post_objects.create(
        title="Post without Read More",
        content="This is the entire content.",
        author=user,
    )
    expected_truncated_content = "<p>This is the entire content.</p>"
    assert post2.truncated_content_markdown == expected_truncated_content


@pytest.mark.django_db
def test_post_is_truncated_property(user):
    # Confirm the truncate tag is set according to settings_testing.py
    truncate_tag = "<!--test-more-->"
    assert settings.TRUNCATE_TAG == truncate_tag

    # Test case 1: Content with truncate tag
    post1 = Post.post_objects.create(
        title="Post with Truncate Tag",
        content=f"This is the intro.{truncate_tag}This is the rest of the content.",
        author=user,
    )
    assert post1.is_truncated is True

    # Test case 2: Content without truncate tag
    post2 = Post.post_objects.create(
        title="Post without Truncate Tag",
        content="This is the entire content.",
        author=user,
    )
    assert post2.is_truncated is False

    # Test case 3: Content with truncate tag at the beginning
    post3 = Post.post_objects.create(
        title="Post with Truncate Tag at the Beginning",
        content=f"{truncate_tag}This is the content.",
        author=user,
    )
    assert post3.is_truncated is True

    # Test case 4: Content with truncate tag at the end
    post4 = Post.post_objects.create(
        title="Post with Truncate Tag at the End",
        content=f"This is the content.{truncate_tag}",
        author=user,
    )
    assert post4.is_truncated is True


@pytest.mark.django_db
def test_post_permalink(user):
    post = Post(
        title="Test Post",
        slug="test-post",
        content="This is a test post.",
        author=user,
        date=timezone.datetime(2024, 1, 1),
        status="published",
        post_type="post",
    )

    # Confirm the post prefix and permalink settings are set according to settings_testing.py
    assert settings.POST_PREFIX == "test-posts"
    assert settings.POST_PERMALINK == ""

    assert post.permalink == "test-posts/test-post"
    settings.set("POST_PERMALINK", settings.DAY_SLUG)
    assert post.permalink == "test-posts/2024/01/01/test-post"
    settings.set("POST_PERMALINK", settings.MONTH_SLUG)
    assert post.permalink == "test-posts/2024/01/test-post"
    settings.set("POST_PERMALINK", settings.YEAR_SLUG)
    assert post.permalink == "test-posts/2024/test-post"

    settings.set("POST_PREFIX", "")
    settings.set("POST_PERMALINK", "")
    assert post.permalink == "test-post"
    settings.set("POST_PERMALINK", settings.DAY_SLUG)
    assert post.permalink == "2024/01/01/test-post"
    settings.set("POST_PERMALINK", settings.MONTH_SLUG)
    assert post.permalink == "2024/01/test-post"
    settings.set("POST_PERMALINK", settings.YEAR_SLUG)
    assert post.permalink == "2024/test-post"

    # Set back to defaults
    settings.set("POST_PREFIX", "test-posts")
    settings.set("POST_PERMALINK", "")


@pytest.mark.django_db
def test_get_published_posts_by_author(user):
    # Create some published posts by the test user
    post1 = Post.post_objects.create(
        title="Post 1",
        content="Content of post 1.",
        author=user,
        status="published",
    )
    post2 = Post.post_objects.create(
        title="Post 2",
        content="Content of post 2.",
        author=user,
        status="published",
    )

    # Create a draft post by the test user
    draft_post = Post.post_objects.create(
        title="Draft Post",
        content="Content of draft post.",
        author=user,
        status="draft",
    )

    # Create a future post by the test user
    future_post = Post.post_objects.create(
        title="Future Post",
        content="Content of future post.",
        author=user,
        status="published",
        date=timezone.now() + timezone.timedelta(days=1),
    )

    # Call the method being tested
    published_posts = Post.post_objects.get_published_posts_by_author(user)

    # Assert that only the published posts by the test user are returned
    assert post1 in published_posts
    assert post2 in published_posts
    assert draft_post not in published_posts
    assert future_post not in published_posts


@pytest.mark.django_db
def test_page_permalink(user):
    page = Post(
        title="Test Page",
        slug="test-page",
        content="This is a test page.",
        author=user,
        date=timezone.now(),
        status="published",
        post_type="page",
    )

    assert page.permalink == "test-page"


@pytest.mark.django_db
def test_get_recent_published_posts(user):
    """Test that the get_recent_published_posts method returns the correct posts."""
    # Confirm settings are set according to settings_testing.py
    assert settings.CACHE_RECENT_PUBLISHED_POSTS is False
    assert settings.RECENT_PUBLISHED_POSTS_COUNT == 3

    # Create some published posts
    post1 = Post.objects.create(title="Post 1", status="published", author=user)
    post2 = Post.objects.create(title="Post 2", status="published", author=user)
    post3 = Post.objects.create(title="Post 3", status="published", author=user)

    # Call the method being tested
    recent_posts = Post.post_objects.get_recent_published_posts()

    # Assert that the correct posts are returned
    assert list(recent_posts) == [post3, post2, post1]

    # Test case 2: Limit the number of posts returned
    settings.set("RECENT_PUBLISHED_POSTS_COUNT", 2)

    assert settings.CACHE_RECENT_PUBLISHED_POSTS is False
    assert settings.RECENT_PUBLISHED_POSTS_COUNT == 2

    # Call the method being tested again
    recent_posts = Post.post_objects.get_recent_published_posts()

    # Assert that the correct posts are returned
    assert list(recent_posts) == [post3, post2]
    assert not post1 in recent_posts


@pytest.mark.django_db
def test_get_published_post_by_path(user):
    """Test that the get_published_post_by_path method returns the correct post."""

    # Confirm settings are set according to settings_testing.py
    assert settings.POST_PREFIX == "test-posts"

    # Create a post
    post = Post.objects.create(title="Test Post", status="published", author=user)

    # Test case 1: POST_PREFIX is set and path starts with POST_PREFIX
    post_path = f"test-posts/{post.slug}"
    assert post == Post.post_objects.get_published_post_by_path(post_path)

    # Test case 2: POST_PREFIX is set but path does not start with POST_PREFIX
    post_path = f"/incorrect-path/{post.slug}"
    # Should raise a ValueError
    with pytest.raises(ValueError):
        Post.post_objects.get_published_post_by_path(post_path)

    # Test case 3: POST_PREFIX is not set but path starts with POST_PREFIX
    settings.set("POST_PREFIX", "")
    post_path = f"test-posts/non-existent-slug"
    # Should raise a ValueError
    with pytest.raises(ValueError):
        Post.post_objects.get_published_post_by_path(post_path)

    # Set back to default
    settings.set("POST_PREFIX", "test-posts")



"""Utility functions that are used in the project."""

import markdown
from django.contrib.auth.models import User
from django.utils.timezone import datetime

from djpress.conf import settings

md = markdown.Markdown(extensions=settings.MARKDOWN_EXTENSIONS, output_format="html")


def render_markdown(markdown_text: str) -> str:
    """Return the Markdown text as HTML."""
    html = md.convert(markdown_text)
    md.reset()

    return html


def get_author_display_name(user: User) -> str:
    """Return the author display name.

    Tries to display the first name and last name if available, otherwise falls back to
    the username.

    Args:
        user: The user.

    Returns:
        str: The author display name.
    """
    if user.first_name and user.last_name:
        return f"{user.first_name} {user.last_name}"

    if user.first_name:
        return user.first_name

    return user.username


def validate_date(year: str, month: str, day: str) -> None:
    """Test the date values.

    Convert the date values to integers and test if they are valid dates.

    The regex that gets the date values checks for the following:
    - year: four digits
    - month: two digits
    - day: two digits

    Args:
        year (str): The year.
        month (str | None): The month.
        day (str | None): The day.

    Raises:
        ValueError: If the date is invalid.

    Returns:
        None
    """
    int_year: int = int(year)
    int_month: int | None = int(month) if month else None
    int_day: int | None = int(day) if day else None

    if int_month == 0 or int_day == 0:
        msg = "Invalid date"
        raise ValueError(msg)

    try:
        if int_month and int_day:
            datetime(int_year, int_month, int_day)

        elif int_month:
            datetime(int_year, int_month, 1)

        else:
            datetime(int_year, 1, 1)

    except ValueError as exc:
        msg = "Invalid date"
        raise ValueError(msg) from exc



"""Post model."""

import logging
from typing import ClassVar

from django.contrib.auth.models import User
from django.core.cache import cache
from django.db import models
from django.utils import timezone
from django.utils.text import slugify

from djpress.conf import settings
from djpress.models import Category
from djpress.utils import render_markdown

logger = logging.getLogger(__name__)


PUBLISHED_POSTS_CACHE_KEY = "published_posts"


class PostsManager(models.Manager):
    """Post custom manager."""

    def get_queryset(self: "PostsManager") -> models.QuerySet:
        """Return the queryset for published posts."""
        return super().get_queryset().filter(post_type="post").order_by("-date")

    def get_recent_published_posts(self: "PostsManager") -> models.QuerySet:
        """Return recent published posts.

        This does not return a paginated queryset. Use get_paginated_published_posts
        instead.

        If CACHE_RECENT_PUBLISHED_POSTS is set to True, we return the cached queryset.
        """
        if settings.CACHE_RECENT_PUBLISHED_POSTS:
            return self._get_cached_recent_published_posts()

        return self.get_published_posts().prefetch_related("categories", "author")[
            : settings.RECENT_PUBLISHED_POSTS_COUNT
        ]

    def get_published_posts(self: "PostsManager") -> models.QuerySet:
        """Returns all published posts.

        For a post to be considered published, it must meet the following requirements:
        - The status must be "published".
        - The date must be less than or equal to the current date/time.
        """
        return self.get_queryset().filter(
            status="published",
            date__lte=timezone.now(),
        )

    def _get_cached_recent_published_posts(self: "PostsManager") -> models.QuerySet:
        """Return the cached recent published posts queryset.

        If there are any future posts, we calculate the seconds until that post, then we
        set the timeout to that number of seconds.
        """
        queryset = cache.get(PUBLISHED_POSTS_CACHE_KEY)

        if queryset is None:
            queryset = (
                self.get_queryset()
                .filter(
                    status="published",
                )
                .prefetch_related("categories", "author")
            )

            timeout = self._get_cache_timeout(queryset)

            queryset = queryset.filter(date__lte=timezone.now())[
                : settings.RECENT_PUBLISHED_POSTS_COUNT
            ]
            cache.set(
                PUBLISHED_POSTS_CACHE_KEY,
                queryset,
                timeout=timeout,
            )

        return queryset

    def _get_cache_timeout(
        self: "PostsManager",
        queryset: models.QuerySet,
    ) -> int | None:
        """Return the timeout for the cache.

        If there are any future posts, we calculate the seconds until that post.

        Args:
            queryset: The queryset of published posts.

        Returns:
            int | None: The number of seconds until the next future post, or None if
            there are no future posts.
        """
        future_posts = queryset.filter(date__gt=timezone.now())
        if future_posts.exists():
            future_post = future_posts[0]
            return (future_post.date - timezone.now()).total_seconds()

        return None

    def get_published_post_by_slug(
        self: "PostsManager",
        slug: str,
    ) -> "Post":
        """Return a single published post.

        Must have a date less than or equal to the current date/time based on its slug.
        """
        # First, try to get the post from the cache
        posts = self.get_recent_published_posts()
        post = next((post for post in posts if post.slug == slug), None)

        # If the post is not found in the cache, fetch it from the database
        if not post:
            try:
                post = self.get_published_posts().get(slug=slug)
            except Post.DoesNotExist as exc:
                msg = "Post not found"
                raise ValueError(msg) from exc

        return post

    def get_published_post_by_path(
        self: "PostsManager",
        path: str,
    ) -> "Post":
        """Return a single published post from a path.

        This is a complex piece of logic...

        Here are all the different valid paths that we need to check:
        - /blog/2021/01/01/post-slug
        - /blog/2021/01/post-slug
        - /blog/2021/post-slug
        - /blog/post-slug
        - /2021/01/01/post-slug
        - /2021/01/post-slug
        - /2021/post-slug
        - /post-slug

        But we could be getting any invalid path too, e.g.
        - /blog/2021/01/01/post-slug/extra
        - /blog/2021/01/post-slug/extra
        - etc.

        This is what we need to do...

        First, we need to know the following:
        - Is there a POST_PREFIX defined?
        - Are DATE_ARCHIVES enabled?
        - If DATE_ARCHIVES are enabled, what is the POST_PERMALINK set to?

        I don't think I can avoid regex matching here...

        For now we'll just look at the POST_PREFIX.
        """
        if settings.POST_PREFIX and path.startswith(settings.POST_PREFIX):
            slug = path.split(settings.POST_PREFIX + "/")[1]
            return self.get_published_post_by_slug(slug)

        if settings.POST_PREFIX and not path.startswith(settings.POST_PREFIX):
            msg = "Invalid path"
            raise ValueError(msg)

        return self.get_published_post_by_slug(path)

    def get_published_posts_by_category(
        self: "PostsManager",
        category: Category,
    ) -> models.QuerySet:
        """Return all published posts for a given category.

        Must have a date less than or equal to the current date/time for a specific
        category, ordered by date in descending order.
        """
        return (
            self.get_published_posts()
            .filter(categories=category)
            .prefetch_related("categories", "author")
        )

    def get_published_posts_by_author(
        self: "PostsManager",
        author: User,
    ) -> models.QuerySet:
        """Return all published posts for a given author.

        Must have a date less than or equal to the current date/time for a specific
        author, ordered by date in descending order.
        """
        return (
            self.get_published_posts()
            .filter(author=author)
            .prefetch_related("categories", "author")
        )


class Post(models.Model):
    """Post model."""

    STATUS_CHOICES: ClassVar = [("draft", "Draft"), ("published", "Published")]
    CONTENT_TYPE_CHOICES: ClassVar = [("post", "Post"), ("page", "Page")]

    title = models.CharField(max_length=200)
    slug = models.SlugField(unique=True, blank=True)
    content = models.TextField()
    author = models.ForeignKey(User, on_delete=models.CASCADE)
    date = models.DateTimeField(default=timezone.now)
    modified_date = models.DateTimeField(auto_now=True)
    status = models.CharField(max_length=10, choices=STATUS_CHOICES, default="draft")
    post_type = models.CharField(
        max_length=10,
        choices=CONTENT_TYPE_CHOICES,
        default="post",
    )
    categories = models.ManyToManyField(Category, blank=True)

    # Managers
    objects = models.Manager()
    post_objects: "PostsManager" = PostsManager()

    class Meta:
        """Meta options for the Post model."""

        verbose_name = "post"
        verbose_name_plural = "posts"

    def __str__(self: "Post") -> str:
        """Return the string representation of the post."""
        return self.title

    def save(self: "Post", *args, **kwargs) -> None:  # noqa: ANN002, ANN003
        """Override the save method to auto-generate the slug."""
        if not self.slug:
            self.slug = slugify(self.title)
            if not self.slug or self.slug.strip("-") == "":
                msg = "Invalid title. Unable to generate a valid slug."
                raise ValueError(msg)
        super().save(*args, **kwargs)

    @property
    def content_markdown(self: "Post") -> str:
        """Return the content as HTML converted from Markdown."""
        return render_markdown(self.content)

    @property
    def truncated_content_markdown(self: "Post") -> str:
        """Return the truncated content as HTML converted from Markdown."""
        read_more_index = self.content.find(settings.TRUNCATE_TAG)
        if read_more_index != -1:
            truncated_content = self.content[:read_more_index]
        else:
            truncated_content = self.content
        return render_markdown(truncated_content)

    @property
    def is_truncated(self: "Post") -> bool:
        """Return whether the content is truncated."""
        return settings.TRUNCATE_TAG in self.content

    @property
    def permalink(self: "Post") -> str:
        """Return the post's permalink.

        The posts permalink is constructed of the following elements:
        - The post prefix - this is configured in POST_PREFIX and could be an empty
          string or a custom string, e.g. `blog` or `posts`.
        - The post date structure - this is configured in POST_PERMALINK and is a
          `strftime` value, e.g. `%Y/%m/%d` or `%Y/%m`. Or it could be an empty string
          to indicate that no date structure is used.
        - The post slug - this is a unique identifier for the post. TODO: should this be
          a database unique constraint, or should we handle it in software instead?
        """
        # We start the permalink with just the slug
        permalink = self.slug

        # If the post type is a page, we return just the slug
        if self.post_type == "page":
            return permalink

        # The only other post type is a post so we don't need to check for that
        # If there's a permalink structure defined, we add that to the permalink
        if settings.POST_PERMALINK:
            permalink = f"{self.date.strftime(settings.POST_PERMALINK)}/{self.slug}"

        # If there's a post prefix defined, we add that to the permalink
        if settings.POST_PREFIX:
            permalink = f"{settings.POST_PREFIX}/{permalink}"

        return permalink



import pytest

from django.urls import reverse

from djpress.conf import settings


def test_url_author_enabled():
    """Test the author URL."""

    # Confirm settings are set according to settings_testing.py
    assert settings.AUTHOR_PATH_ENABLED is True
    assert settings.AUTHOR_PATH == "test-url-author"

    author_url = reverse("djpress:author_posts", args=["test-author"])

    assert author_url == f"/{settings.AUTHOR_PATH}/test-author/"


def test_url_category_enabled():
    """Test the category URL."""

    # Confirm settings are set according to settings_testing.py
    assert settings.CATEGORY_PATH_ENABLED is True
    assert settings.CATEGORY_PATH == "test-url-category"

    category_url = reverse("djpress:category_posts", args=["test-category"])

    assert category_url == f"/{settings.CATEGORY_PATH}/test-category/"
